#include "CqSOM2d.h"

// with SSE this class just works with sizes which are %4==0 because of memory alignement

/*inline float sqrt_SSE(const float f){
	__m128 s;

	s=_mm_set_ps1(f);
	_mm_sqrt_ps(s);

	return s.m128_f32[0];
}

inline float mul_SSE(const float f1,const float f2){
	__m128 s1,s2;

	s1=_mm_set_ps1(f1);
	s2=_mm_set_ps1(f2);
	s1 = _mm_mul_ps(s1,s2);

	return s1.m128_f32[0];
}*/

qSOM2d :: qSOM2d(){
	char szTemp[] = "LampelSoft SOM Simulator 0.2 by Johannes Lampel\n";
	m_szVersion = new char[strlen(szTemp)+1];
	strcpy(m_szVersion,szTemp);
	
	m_dp3Weights	= 0;
	m_dp1Input	= 0;
	m_pfDistFuncL = 0;
	m_pVDiff		= 0;
	m_dLearnrate	= 0.0;
	m_dDistanceP	= 0.0;
	m_lCat = 0;

	// init pointer to Neightborhoodkernel function
	m_fNK = fNKgauss1;

	// INIT RAND
	srand(time(0));
}

qSOM2d :: ~qSOM2d(){
#ifdef __SSE
	if(m_dp3Weights)	_aligned_free(m_dp3Weights);
	if(m_dp1Input)		_aligned_free(m_dp1Input);
#else
	if(m_dp3Weights)	delete [] m_dp3Weights;
	if(m_dp1Input)		delete [] m_dp1Input;
#endif
	if(m_szVersion)		delete [] m_szVersion;
	if(m_pVDiff)		delete m_pVDiff;

	m_dp3Weights	= 0;
	m_dp1Input	= 0;
	m_szVersion	= 0;
	m_pVDiff		= 0;
}

int qSOM2d :: SetSize(long lINeurons,long lpXSize,long lpYSize){
	m_lNNeurons	= lINeurons;
#ifdef __SSE
	m_lNNeuronsA = m_lNNeurons%4?(m_lNNeurons/4+1)*4:m_lNNeurons;
	cout << m_lNNeurons << " - " << m_lNNeuronsA <<endl;
#else
	m_lNNeuronsA = m_lNNeurons;
#endif
	m_lXSize	= lpXSize;
	m_lYSize	= lpYSize;
	m_lRow		= m_lXSize * m_lNNeuronsA;
	
	// alloc mem for weights - 3dim
	// usage : weigth := weights[ly * m_lXSize * m_lNNeuronsA + lx * m_lNNeuronsA + lN]
	// m_lXSize * m_lNNeuronsA := m_lRow
	
#ifdef __SSE
	if(m_dp3Weights) _aligned_free(m_dp3Weights);
	m_dp3Weights = (nVec_var*)_aligned_malloc(sizeof(nVec_var)*(m_lNNeuronsA * m_lXSize * m_lYSize/4+1)*4,16);
#else
	if(m_dp3Weights) delete [] m_dp3Weights;
	m_dp3Weights = new nVec_var[m_lNNeuronsA * m_lXSize * m_lYSize];
#endif
	
	// alloc mem for input neurons
#ifdef __SSE
	if(m_dp1Input) _aligned_free( m_dp1Input );
	m_dp1Input = (nVec_var*)_aligned_malloc(sizeof(nVec_var)*m_lNNeuronsA,16);
#else	
	if(m_dp1Input) delete [] m_dp1Input;
	m_dp1Input = new nVec_var [m_lNNeuronsA];
#endif

	// Alloc space for VDiff - Diff between input and winner vector

	if(m_pVDiff) delete m_pVDiff;
	m_pVDiff = new nVec(m_lNNeurons);

	// alloc space for pfDistFuncL
#ifdef __SSE
	if(m_pfDistFuncL) _aligned_free(m_pfDistFuncL);
	m_pfDistFuncL = (float *)_aligned_malloc( sizeof(nVec_var)*m_lXSize * m_lYSize,16);
#else
	if(m_pfDistFuncL) delete [] m_pfDistFuncL;
	m_pfDistFuncL = new nVec_var[m_lXSize * m_lYSize];
#endif
	
	return 1;
}

int qSOM2d :: InitWeights(InitType IType){
	long lschl;
	switch (IType){
	case Init_Random:
		for(lschl=0;lschl < m_lNNeuronsA * m_lXSize * m_lYSize;lschl++){
			m_dp3Weights[lschl] = nVec_var(rand()) / RAND_MAX;
		}
		return 0;
	default:
		cerr << "\"What's that\" - InitWeights"<<endl;
		return 1;
	}
}

int qSOM2d :: SetInput(nVec_var *dptr){
	memcpy(m_dp1Input,dptr,sizeof(nVec_var) * m_lNNeurons);
	return 0;
}

void qSOM2d :: GetWinner(long &lxMin,long &lyMin){			// GetNearest Neuron to Input
	try{
		if(m_dp1Input){
			// set vars to 0
			// this has to be done, 'cause sometimes ( or only with optimized compilin' with MS C++6.0 he cannot find the nearest Neuron !!!??) this function isn't able to find minima and would otherwise cause a page fault
			lyMin = lxMin = 0;
			
			long lxschl,lyschl;
			m_dWinnerMin = FLT_MAX;	// dist of best neuron
			nVec_var dLength;
			long lsP;
			
			nVec VInput (m_lNNeurons);
			nVec VComp	(m_lNNeurons);
			nVec VDiff	(m_lNNeurons);
			
			VInput.SetMemVec(m_dp1Input);			// don't copy data - only set pointer
			
			for(lyschl = 0;lyschl < m_lYSize;lyschl++){
				lsP = lyschl * m_lRow;
				for(lxschl = 0;lxschl < m_lXSize;lxschl++){
					VComp.GetVec(&m_dp3Weights[lsP + lxschl * m_lNNeuronsA]);
					VComp -= VInput;
					dLength = VComp.Length();
					//dLength = VComp.CBLength();
					//dLength = VComp.getMin();
					
					if(dLength < m_dWinnerMin){
						m_dWinnerMin = dLength;
						lxMin = lxschl;
						lyMin = lyschl;
					}
				}
			}
		}
		else{
			throw NoInput();
		}
	}
	catch(NoInput NI){
		cerr << "No Input specified" << endl;
		throw NI;
	}
}

nVec_var qSOM2d :: Categorize(nVec_var *dpParam){
	SetInput(dpParam);
	return Categorize();
}

#ifdef __SSE
__m128 m_fNKcone_SSE(__m128 dist,float distp){
	__m128 temp;
	__m128 m128distp;

	temp = _mm_set_ps1(1);
	m128distp = _mm_set_ps1(distp);
	dist = _mm_mul_ps(dist,m128distp);
	temp = _mm_sub_ps(temp,dist);

	long lxt;
	for(lxt=0;lxt < 4 ; lxt ++){
		if(temp.m128_f32[lxt] < 0.f)
			temp.m128_f32[lxt] = 0.f;
	}

	return temp;
}
#endif

nVec_var qSOM2d :: Categorize(void){		// learning function
	m_lCat ++;
	long lxWinner=0,lyWinner=0;
	long lxschl,lyschl;
	nVec_var ddistance,dD = 1.f/m_dDistanceP;

	long lDx,lDy;
	nVec	VIn(m_lNNeurons);			// input vector
	nVec	VNeuron(m_lNNeurons);		// vector of current neuron
	//nVec	VUpdated(m_lNNeurons);	// updated vector
	nVec	VTemp(m_lNNeurons);		// temp
	
	GetWinner(lxWinner,lyWinner);							// getting winner neuron
	VIn.GetVec(m_dp1Input);									// get input to vector

	m_pVDiff->GetVec(GetNWeights(lxWinner,lyWinner));
	*m_pVDiff -= VIn;

#ifdef __SSE
	__m128 m128dist,m128distf,m128dlrt;
	m128dlrt = _mm_set_ps1(m_dLearnrate);
#endif
	
	long lTemp,lxt;
#ifdef __SSE
	for(lyschl=0;lyschl < m_lYSize;lyschl++){
		lTemp = lyschl * m_lRow;
		lDy = lyschl - lyWinner;
		lDy *= lDy;
		for(lxschl=0;lxschl < m_lXSize;lxschl+=4){
			for(lxt = 0; lxt < 4; lxt ++){
				// get distance to winner
				lDx = lxschl + lxt - lxWinner;
				m128dist.m128_f32[lxt] = (float)(lDx * lDx + lDy);
			}
			m128dist = _mm_sqrt_ps(m128dist);
			for(lxt = 0; lxt < 4; lxt ++){
				m128distf.m128_f32[lxt] = m_fNK(m128dist.m128_f32[lxt],m_dDistanceP);
			}
			//m128distf = m_fNKcone_SSE(m128dist,dD);
			m128distf = _mm_mul_ps(m128distf,m128dlrt);
			*((__m128 *)&(m_pfDistFuncL[lyschl * m_lXSize + lxschl])) = m128distf;
		}
	}
#else
	for(lyschl=0;lyschl < m_lYSize;lyschl++){
		lTemp = lyschl * m_lRow;
		lDy = lyschl - lyWinner;
		lDy *= lDy;
		for(lxschl=0;lxschl < m_lXSize;lxschl++){
			// get distance to winner
			lDx = lxschl - lxWinner;
			ddistance = sqrt(lDx * lDx + lDy);
			pfDistFuncL[lyschl * m_lXSize + lxschl] = m_fNK(ddistance,m_dDistanceP)*m_dLearnrate;
		}
	}
#endif
	for(lyschl=0;lyschl < m_lYSize;lyschl++){
		lTemp = lyschl * m_lRow;
		for(lxschl=0;lxschl < m_lXSize;lxschl++){
			// get current weights
			VNeuron.SetMemVec(&m_dp3Weights[lTemp + lxschl * m_lNNeuronsA]);
			// update weights
			
			VTemp = VIn;
			VTemp -= VNeuron;
			VTemp *= m_pfDistFuncL[lyschl * m_lXSize + lxschl];
			VNeuron += VTemp;
			// don't have to write updated vector back to mem cause it was calced in mem (VNeuron.SetMemVec)
		}
	}
/*	// this is the straight way, unlucky for SSE i though. As it turned out, the solution above is even faster for non SSE use

	nVec_var fDistFuncL;
	for(lyschl=0;lyschl < m_lYSize;lyschl++){
		lTemp = lyschl * m_lRow;
		for(lxschl=0;lxschl < m_lXSize;lxschl++){
			// get current weights
			dpCVec = &m_dp3Weights[lTemp + lxschl * m_lNNeurons];
			// get distance to winner
			lDx = lxschl - lxWinner;
			lDy = lyschl - lyWinner;
			ddistance = sqrt(lDx * lDx + lDy * lDy);
			fDistFuncL = m_fNK(ddistance,m_dDistanceP) * m_dLearnrate;
			// get weights from current neuron
			VNeuron.SetMemVec(dpCVec);
			// update weights
			
			VTemp = VIn;
			VTemp -= VNeuron;
			VTemp *= fDistFuncL;
			VNeuron += VTemp;
			// don't have to write updated vector back to mem cause it was calced in mem (VNeuron.SetMemVec)
		}
	}*/
	return m_pVDiff->Length();
}

nVec_var qSOM2d :: CategorizeFast(nVec_var *dpParam){
	SetInput(dpParam);
	return CategorizeFast();
}

nVec_var qSOM2d :: CategorizeFast(void){
/* fast learning function -> not suitable for gauss1 + 2  -> only for cone,cyl and cos because for the gauss' the result of them isn't 0 when z > d
This function neglects the neurons outside a square defined by the winner neuron as middle and the min radius by the 'm_dDistanceP' The distance or stiffness param
*/

	long lxWinner,lyWinner;
	long lxschl,lyschl;
	nVec_var ddistance;
	long lDx,lDy;
	nVec	VIn(m_lNNeurons);			// input vector
	nVec	VNeuron(m_lNNeurons);		// vector of current neuron
	nVec	VUpdated(m_lNNeurons);	// updated vector
	nVec	VTemp(m_lNNeurons);		// temp
	nVec_var	*dpCVec;
	long lxstart,lxend,lystart,lyend;
	
	GetWinner(lxWinner,lyWinner);							// getting winner neuron
	VIn.GetVec(m_dp1Input);									// get input to vector

	m_pVDiff->GetVec(GetNWeights(lxWinner,lyWinner));
	*m_pVDiff -= VIn;

	// calc vars for only changing weights in neurons which are in the radius of m_dDistanceP
	if(m_dDistanceP < 1.0){					// no other neurons are affected
		lxstart = lxWinner;
		lxend = lxstart + 1;
		lystart = lyWinner;
		lyend = lystart + 1;
	}
	else{									// otherwise test ...
		lxstart = long(lxWinner - m_dDistanceP - 1);
		lxend	= long(lxWinner + m_dDistanceP + 1);
		lystart = long(lyWinner - m_dDistanceP - 1);
		lyend	= long(lyWinner + m_dDistanceP + 1);
		if(lxstart < 1){					// if outside of the borders
			lxstart = 0;
		}
		if(lystart < 1){
			lystart = 0;
		}
		if(lxend > m_lXSize){
			lxend = m_lXSize;
		}
		if(lyend > m_lYSize){
			lyend = m_lYSize;
		}
	}
	
	for(lyschl=lystart;lyschl < lyend;lyschl++){
		for(lxschl=lxstart;lxschl < lxend;lxschl++){
			// get distance to winner
			lDx = lxschl - lxWinner;
			lDy = lyschl - lyWinner;
			ddistance = sqrt(lDx * lDx + lDy * lDy);
			m_pfDistFuncL[lxschl * m_lYSize + lyschl] = m_dLearnrate * m_fNK(ddistance,m_dDistanceP);
		}
	}
	for(lyschl=lystart;lyschl < lyend;lyschl++){
		for(lxschl=lxstart;lxschl < lxend;lxschl++){
			// get current weights
			dpCVec = &m_dp3Weights[lyschl * m_lRow + lxschl * m_lNNeuronsA];
			// get weights from current neuron
			VNeuron.SetMemVec(dpCVec);
			// update weights
			VTemp = VIn;
			VTemp -= VNeuron;
			VTemp *= m_pfDistFuncL[lxschl * m_lYSize + lyschl];
			VNeuron += VTemp;
			// don't have to write updated vector back to mem cause it was calced in mem (VNeuron.SetMemVec)
		}
	}
	return m_pVDiff->Length();
}

nVec_var *qSOM2d :: GetNWeights(long lx,long ly){
	try{
		if(lx < 0 ||ly < 0)
			throw FuckinParam();
		return &m_dp3Weights[ly * m_lRow + lx * m_lNNeuronsA];
	}
	catch(FuckinParam FP){
		cerr << "Got some fuckin' params in qSOM2d :: GetNWeights"<<endl;
		throw FP;
	}
}

nVec_var qSOM2d :: GetDistP(void){
	return m_dDistanceP;
}

void qSOM2d :: SetDistP(nVec_var dParam){
	m_dDistanceP = dParam;
}

nVec_var qSOM2d :: GetLRate(void){
	return m_dLearnrate;
}

void qSOM2d :: SetLRate(nVec_var dParam){
	m_dLearnrate = dParam;
}

void qSOM2d :: Save(const char *szFileName){
	FILE *fhd;
	long lCoding = _SOM_ID;
	try{
		if((fhd = fopen(szFileName,"wb"))){
			fwrite(&lCoding,sizeof(long),1,fhd);			// writing id for somfiles
			fwrite(&m_lXSize,sizeof(long),1,fhd);
			fwrite(&m_lYSize,sizeof(long),1,fhd);
			fwrite(&m_lNNeurons,sizeof(long),1,fhd);
			fwrite(&m_lNNeuronsA,sizeof(long),1,fhd);
			fwrite(&m_lRow,sizeof(long),1,fhd);
			fwrite(&m_dLearnrate,sizeof(nVec_var),1,fhd);
			fwrite(&m_dDistanceP,sizeof(nVec_var),1,fhd);
			fwrite(m_dp3Weights,sizeof(nVec_var),m_lXSize * m_lYSize * m_lNNeurons,fhd);
			fclose(fhd);
		}
		else
		{
			throw FileSaveError();
		}
	}
	catch(FileSaveError FSE){
		cerr << "Error opening file : " << szFileName << endl;
		throw FSE;
	}
}

void qSOM2d :: Load (const char *szFileName){
	FILE *fhd;
	long lCoding;
	try{
		if((fhd=fopen(szFileName,"rb"))){
#ifdef __SSE
			if(m_dp3Weights) _aligned_free(m_dp3Weights);
			if(m_dp1Input) _aligned_free( m_dp1Input );
#else
			if(m_dp3Weights)	delete [] m_dp3Weights;
			if(m_dp1Input)	delete [] m_dp1Input;
#endif
			
			fread(&lCoding,sizeof(long),1,fhd);					// read coding and verify
			if(lCoding != _SOM_ID)
				throw UnknownFileFormat();
			fread(&m_lXSize,sizeof(long),1,fhd);					// just as save - vice versa
			fread(&m_lYSize,sizeof(long),1,fhd);
			fread(&m_lNNeurons,sizeof(long),1,fhd);
			fread(&m_lNNeuronsA,sizeof(long),1,fhd);
			fread(&m_lRow,sizeof(long),1,fhd);
			
			fread(&m_dLearnrate,sizeof(nVec_var),1,fhd);
			fread(&m_dDistanceP,sizeof(nVec_var),1,fhd);
			
#ifdef __SSE
			m_dp3Weights = (nVec_var*)_aligned_malloc(sizeof(nVec_var)*(m_lNNeuronsA * m_lXSize * m_lYSize/4+1)*4,16);
			m_dp1Input = (nVec_var*)_aligned_malloc(sizeof(nVec_var)*m_lNNeuronsA,16);
#else
			m_dp3Weights	= new nVec_var [m_lXSize * m_lYSize * m_lNNeuronsA];
			m_dp1Input	= new nVec_var [m_lNNeuronsA];
#endif
#ifdef __SSE
			if(m_pfDistFuncL) _aligned_free(m_pfDistFuncL);
			m_pfDistFuncL = (float *)_aligned_malloc( sizeof(nVec_var)*m_lXSize * m_lYSize,16);
#else
			if(m_pfDistFuncL) delete [] m_pfDistFuncL;
			m_pfDistFuncL = new nVec_var[m_lXSize * m_lYSize];
#endif
			
			fread(m_dp3Weights,sizeof(nVec_var),m_lXSize * m_lYSize * m_lNNeuronsA,fhd);
			memset(m_dp1Input,0,sizeof(nVec_var) * m_lNNeuronsA);
			fclose(fhd);
		}
		else
		{
			throw FileLoadError();
		}
	}
	catch(FileLoadError FLE){
		cerr << "Error opening file : " << szFileName << endl;
		throw FLE;
	}
	catch(UnknownFileFormat UFF){
		cerr << "Unknown file format" << endl;
		throw UFF;
	}
}

void qSOM2d :: SetNKFunction(nVec_var (*m_fNKParam)(const nVec_var&,const nVec_var&)){
	m_fNK = m_fNKParam;
}

void *qSOM2d :: GetNKFunction(void){
	return ((void *)(m_fNK));
}
