#ifndef __CQSOM2D_H
#define __CQSOM2D_H

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/

// class for 2d SOMs
// by Johannes Lampel / LampelSoft
// (c) Johannes Lampel / LampelSoft
// written Sept 13rd 2000

#include "GlobalCqSOMDefs.h"
#include "nVec.h"
#include "nSOMErr.h"
#include "fNBorKernel.h"

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <iostream.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <float.h>

#define _SOM_ID 0x20000927					// ID for som - files

class DllExport qSOM2d{
public:
	qSOM2d();									// ctor
	~qSOM2d();									// dtor

	int SetSize(long,long,long);				// set size of map (Dim of Input,x,y)	-> x=y isn't too useful
	int InitWeights(InitType);					// ...
	int SetInput(nVec_var *);						// copies input into local class storage
	nVec_var Categorize(void);						// cats input
	nVec_var Categorize(nVec_var*);					// like the two above
	nVec_var CategorizeFast(void);					// cats faster cause it neglects some (!) neurons
	nVec_var CategorizeFast(nVec_var*);				// ...
	void GetWinner(long &,long &);				// finds 'nearest' neuron
	nVec_var *GetNWeights(long,long);				// return the weight matrix of one neuron
	nVec_var GetDistP(void);						// return stiffness param
	void SetDistP(nVec_var);						// ..
	nVec_var GetLRate(void);						// return lRate
	void SetLRate(nVec_var);						// ..
	void SetNKFunction(nVec_var (*name)(const nVec_var&,const nVec_var&));	// sets the pointer to the NK functions, which is set to fNKgauss1 by default
	void *GetNKFunction(void);					// ..
	void Save(const char *);					// savin' data to file
	void Load(const char *);					// loadin' data from file to mem

	int getDimensionality(void){return m_lNNeurons;}
	int getXSize(void){return m_lXSize;}
	int getYSize(void){return m_lYSize;}
	nVec_var getWinnerMin(void){return m_dWinnerMin;}

	long m_lXSize,m_lYSize;
protected:
	long m_lNNeuronsA;
	long m_lRow;									// lXSize * lNNeurons	(just performance)
	nVec_var *m_dp3Weights;							// field of weights
	nVec_var *m_dp1Input;							// input to be categorized or tested
	nVec_var m_dLearnrate;							// ...
	nVec_var m_dDistanceP;							// stiffness param
	nVec_var m_dWinnerMin;
	nVec_var (*m_fNK)(const nVec_var&,const nVec_var&); // pointer to NK function
	nVec	*m_pVDiff;							// diff (updated by calling cat) between Input and Winner
	nVec_var *m_pfDistFuncL; // field containing distance function results multiplied with learnrate
	long m_lNNeurons;
	long m_lCat;
	char *m_szVersion;							// Just some information about this simulator
private:
};

#endif  __CQSOM2D_H