#ifndef __NVEC_H
#define __NVEC_H

// n-dim vectorclass
// by Johannes Lampel / LampelSoft
// (c)  copyright Johannes Lampel
// written Sept 13rd 2000

#include "GlobalCqSOMDefs.h"
#include "nVecErr.h"
#include <stdio.h>
#include <limits.h>
#include <math.h>
#include <memory.h>
#include <iostream.h>

typedef float nVec_var ;

#define __SSE

#ifdef __SSE
#include <xmmintrin.h>
#endif

enum nVecTypes{
	nVT_copy=0,
	nVT_mem
};

class nVec;

nVec operator *(nVec_var,nVec&);

class DllExport nVec{
	friend nVec operator *(nVec_var,nVec&);
	friend class SOMPattern;
public:
	nVec()
	{
		lCDim = 0;
		iType = -1;
		ppdDim = 0;
	}
	nVec(const dim_c dcDim)
	{
#ifndef __SSE
		ppdDim = new nVec_var[(dcDim/4+1)*4];
#else

		ppdDim = (nVec_var*) _aligned_malloc( /*dcDim%4?dcDim:*/((dcDim/4+1)*4) * sizeof(nVec_var), 16);
#endif
		lCDim = dcDim;
		iType = nVT_copy;							// copy data by default
	}
	nVec(const dim_c dcDim,const nVec_var * const dptr)
	{
#ifndef __SSE
		ppdDim = new nVec_var[(dcDim/4+1)*4];
#else
		ppdDim = (nVec_var*) _aligned_malloc( /*dcDim%4?dcDim:*/((dcDim/4+1)*4) * sizeof(nVec_var), 16);
#endif
		lCDim = dcDim;
		memcpy(ppdDim,dptr,sizeof(nVec_var) * lCDim);
		iType = nVT_copy;							// copy data by default
	}
	nVec(const nVec &npVec){
		ppdDim = 0;
		this->operator = (npVec);
		iType = nVT_copy;							// copy data by default
	}

	virtual ~nVec();

	nVec_var	Length(void)							// ...
	{
		return sqrt(SLength());
	}

	int		Normalize(void)						// ...
	{
		nVec_var dLength = Length();
		if(dLength==0.0){
			dLength=1.0;
			return 1;
		}
		dLength = 1.f / dLength;

		this->operator *=(dLength);
		return 1;
	}

	int		GetVec(nVec_var *dptr)						// copy data from field into this class
	{
		memcpy(ppdDim,dptr,sizeof(nVec_var) * lCDim);
		return 1;
	}

	int		SetMemVec(nVec_var *dpParam)					// set ddpDim to Param and iType to nVT_mem - be careful, you havn't a copy of this data, your changing the original data
	{
		// setting ppdDim to mempos and iType to nVT_mem (case del of old alloc mem)
		if(!iType){
#ifdef __SSE
			if(ppdDim) _aligned_free ( ppdDim );;
#else
			if(ppdDim) delete [] ( ppdDim );;
#endif
		}
		ppdDim = dpParam;
		iType = nVT_mem;
		return 1;
	}

	int		ConvMemtoCopy(void)						// convert a Mem - typed instance of this class to a instance, which contains a copy of this data items - alloc mem and copy
	{
		nVec_var *dpTmp;
#ifdef __SSE
		dpTmp = (nVec_var*) _aligned_malloc((lCDim/4+1)*4 * sizeof(nVec_var), 16);
#else
		dpTmp = new nVec_var[(lCDim/4+1)*4];
#endif
		memcpy(dpTmp,ppdDim,sizeof(nVec_var) * lCDim);
		ppdDim = dpTmp;
		iType = nVT_copy;
		return 1;
	}

	int		SetVec(nVec_var *dpDest)						// copy data from class to field
	{
		memcpy(dpDest,ppdDim,sizeof(nVec_var) * lCDim);
		return 1;
	}
	nVec_var	SLength(void)							// return sum of all quadrs of comps
	{
#ifndef SHIT			// normal is faster
		nVec_var dQSum = 0,dTemp;									// sum of quadr
		long lschl;
		for(lschl = 0;lschl < lCDim;lschl ++){
			dTemp = ppdDim[lschl];
			dQSum += dTemp * dTemp;
		}
		return dQSum;
#else
		long lSize = lCDim / 4,lschl;

		for(lschl = lSize*4;lschl < lCDim;lschl ++){		// zero extra aligned data ( cause that's not officially belonging to this vector )
			ppdDim[lschl] = 0.f;
		}

		lSize ++ ; // "+1 because we just loop down to 1"

		__m128 *add1 = (__m128*)ppdDim;
		__m128 sum;

		sum = _mm_setzero_ps();

		for(; lSize; lSize--){
			sum = _mm_add_ps(sum,_mm_mul_ps(*add1,*add1));

			add1++;
		}

		return  sum.m128_f32[0] + sum.m128_f32[1] + sum.m128_f32[2] + sum.m128_f32[3];
#endif
	}

	nVec_var getMax(void){
		nVec_var max = -1000000000;
		long lschl;
		for(lschl = 0;lschl < lCDim;lschl ++){
			if(ppdDim[lschl] > max){
				max = ppdDim[lschl];
			}
		}
		return max;
	}
	nVec_var getMin(void){
		nVec_var min = 1000000000;
		long lschl;
		for(lschl = 0;lschl < lCDim;lschl ++){
			if(ppdDim[lschl] < min){
				min = ppdDim[lschl];
			}
		}
		return min;
	}

	nVec_var	CBLength(void);							// return CBDist
	void	Zero(void)
	{
#ifndef __SSE
		long lschl;
		for(lschl=0;lschl < lCDim;lschl++){
			ppdDim[lschl] = 0;
		}
#else
		__m128 m128_zero;
		long lSize = lCDim / 4 + 1;		// +1 because we just loop down to 1
		__m128 *dest = (__m128*)ppdDim;

		m128_zero = _mm_set_ps1(0);

		for(; lSize; lSize--){
			*dest = m128_zero;
			dest ++;
		}
#endif
	}

	void	ZeroAligned(void){
		long lschl;		// set last components to zero ...

		long lSize = lCDim / 4;
		for(lschl = lSize*4;lschl < lCDim;lschl ++){
			ppdDim[lschl] = 0.f;
		}
	}

	int		SetNumDim(const dim_c dcDim){
#ifdef __SSE
		if(!iType && !ppdDim) _aligned_free ( ppdDim );;
		lCDim = dcDim;
		ppdDim = ppdDim = (nVec_var*) _aligned_malloc((dcDim/4+1)*4 * sizeof(nVec_var), 16);
#else
		if(!iType && !ppdDim) delete [] ( ppdDim );;
		lCDim = dcDim;
		ppdDim = new nVec_var[(dcDim/4+1)*4];
#endif
		return 1;
	}

	int		Load(FILE *);
	int		Save(FILE *);

	void testout(void);

	const nVec & operator = (const nVec &NParam)	// ...
	{
		try{
			if(!iType){
#ifdef __SSE
				if(ppdDim) _aligned_free(ppdDim);
				ppdDim = (nVec_var*) _aligned_malloc((NParam.lCDim/4+1)*4 * sizeof(nVec_var), 16);
#else
				if(ppdDim) delete [] ppdDim;
				ppdDim = new nVec_var[(NParam.lCDim/4+1)*4];
#endif
				lCDim = NParam.lCDim;
#ifndef SHIT		// fast the old way
				memcpy(ppdDim,NParam.ppdDim,sizeof(nVec_var)*lCDim);
#else
				// this is slower using SSE
				long lSize = lCDim / 4 + 1;		// +1 because we just loop down to 1
				__m128 *dest = (__m128*)ppdDim;
				__m128 *src = (__m128*)NParam.ppdDim;

				for(; lSize; lSize--){
					_mm_store_ps ( (float *)dest,_mm_load_ps((float *)src));
					dest ++;
					src ++;
				}
#endif
			}
			else{
				if(lCDim != NParam.lCDim){
					throw IncompatibleVectors();
				}
				memcpy(ppdDim,NParam.ppdDim,sizeof(nVec_var)*lCDim);
			}
			return ((const nVec &)(*this));
		}
		catch(IncompatibleVectors IV){
			cerr << " 'IncompatibleVectors' thrown by = " << lCDim << " + " << NParam.lCDim <<" : "<< &NParam <<endl;
			throw (IV);
		}
	}
	const nVec & operator +=(const nVec &NParam)	// ...
	{
		try{
			if(lCDim != NParam.lCDim){
				throw IncompatibleVectors();
			}
#ifndef __SSE
			long lschl;
			for(lschl=0;lschl<lCDim;lschl++){
				ppdDim[lschl] += NParam.ppdDim[lschl];
			}
#else
			long lSize = lCDim / 4 + 1;		// +1 because we just loop down to 1

			__m128 *add1 = (__m128*)ppdDim;
			__m128 *add2 = (__m128*)NParam.ppdDim;

			for(; lSize; lSize--){
				*add1 = _mm_add_ps(*add1,*add2);

				add1++;
				add2++;
			}
#endif
		}
		catch(IncompatibleVectors IV){
			cerr << " 'IncompatibleVectors' thrown by += " << lCDim << " + " << NParam.lCDim <<" : "<< &NParam <<endl;
			throw(IV);
		}
		return ((const nVec &)(*this));
	}
	const nVec & operator -=(const nVec &NParam)	// ...
	{
		try{
			if(lCDim != NParam.lCDim){
				throw IncompatibleVectors();
			}
#ifndef __SSE
			long lschl;
			for(lschl=0;lschl < lCDim;lschl++){
				ppdDim[lschl] -= NParam.ppdDim[lschl];
			}
#else
			long lSize = lCDim / 4 + 1;		// +1 because we just loop down to 1

			__m128 *sub1 = (__m128*)ppdDim;
			__m128 *sub2 = (__m128*)NParam.ppdDim;
			__m128 dest;

			for(; lSize; lSize--){
				dest = _mm_sub_ps(*sub1,*sub2);
				*sub1 = dest;

				sub1++;
				sub2++;
			}
#endif
		}
		catch(IncompatibleVectors IV){
			cerr << " 'IncompatibleVectors' thrown by -= " << lCDim << " + " << NParam.lCDim <<" : "<< &NParam <<endl;
			throw(IV);
		}
		return ((const nVec &)(*this));
	}
	const nVec & operator *=(const nVec_var dParam)	// ...
	{
#ifndef __SSE
		long lschl;
		for(lschl=0;lschl<lCDim;lschl++){
			ppdDim[lschl] *= dParam;
		}
#else
		__m128 m128_fac;

		m128_fac = _mm_set_ps1(dParam);

		long lSize = lCDim / 4 + 1;		// +1 because we just loop down to 1

		__m128 *dest = (__m128*)ppdDim;

		for(; lSize; lSize--){
			*dest = _mm_mul_ps(*dest,m128_fac);
			dest ++;
		}
#endif
		return ((const nVec &)(*this));
	}
	const nVec & operator /=(const nVec_var dParam)	// ...
	{
		nVec_var fFac = 1.f / dParam;

		this->operator *= (fFac);

		return ((const nVec &)(*this));
	}

	nVec		operator +(const nVec &);	// ...
	nVec		operator -(const nVec &);	// ...
	nVec		operator *(const nVec_var);	// ...
	nVec		operator /(const nVec_var);	// ...

	nVec_var &	operator [](long lIndex)			// return value of [dimension] component
	{
		return ((nVec_var &)(ppdDim[lIndex]));
	}
	nVec_var *getData(void){return ppdDim;}
protected:
	nVec_var *ppdDim;									// data items
	long lCDim;										// number of dimensions
	int iType;										// copy data or set pointer to data
private:
};

#endif  __NVEC_H