// Winamp test visualization library v1.0
// Copyright (C) 1997-1998, Justin Frankel/Nullsoft
// Feel free to base any plugins on this "framework"...

#include <windows.h>		// Header File For Windows
#include <math.h>
#include <gl\gl.h>			// Header File For The OpenGL32 Library
#include <gl\glu.h>			// Header File For The GLu32 Library
#include <gl\glaux.h>		// Header File For The Glaux Library
#include <stdio.h>
#include <time.h>

#include "../som/som.h"

#include "cframetimer.h"

CFrameTimer CFTimer;

#include "../size.h"

#include "vis.h"
#include <list>
using namespace std;

#define _MAX_LISTS 800
#define _MAX_WINNERDIST 400

#define _WINNERDISTDS 9.f

char szAppName[] = "SimpleVis"; // Our window class, etc

// configuration declarations
int config_x=50, config_y=50;	// screen X position and Y position, repsectively
void config_read(struct winampVisModule *this_mod);		// reads the configuration
void config_write(struct winampVisModule *this_mod);	// writes the configuration
void config_getinifn(struct winampVisModule *this_mod, char *ini_file); // makes the .ini file filename

// returns a winampVisModule when requested. Used in hdr, below
winampVisModule *getModule(int which);

// "member" functions
void config(struct winampVisModule *this_mod); // configuration dialog
int init(struct winampVisModule *this_mod);	   // initialization for module
int render1(struct winampVisModule *this_mod);  // rendering for module 1
int render2(struct winampVisModule *this_mod);  // rendering for module 2
int render3(struct winampVisModule *this_mod);  // rendering for module 3
void quit(struct winampVisModule *this_mod);   // deinitialization for module

// our window procedure
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
HWND hMainWnd; // main window handle

//static SOMPattern SP(_DIM);
static qSOM2d gSOM;
nVec_var dpInput[_DIM*(_PATTERNPERINPUT+1)];
static float pfOldData[_DIM * (_PATTERNPERINPUT+1)];

static long lFrame = 0;

static bool g_bStatic = false;
static int g_iFInsts = 1;
static float g_fzOffset = 2;

static float pdLT1[_X][_Y];
static float pdLT2[_X][_Y];
static float pdLT3[_X][_Y];

static float pppfVertex2[_X][_Y][3];
static float pppfColor2[_X][_Y][3];

bool bwires = false;
bool bDisplayFPS = true;
bool bSwitchMode = false;

list <nVec_var> g_LWinnerDist;

// Double buffering data
/*HDC memDC;		// memory device context
HBITMAP	memBM,  // memory bitmap (for memDC)
oldBM;  // old bitmap (from memDC)*/

GLuint dplists;
GLuint dplists1;
int iNow=0;

HDC			hDC=NULL;		// Private GDI Device Context
HGLRC		hRC=NULL;		// Permanent Rendering Context
HWND		hWnd=NULL;		// Holds Our Window Handle
HINSTANCE	hInstance;		// Holds The Instance Of The Application

bool	keys[256];			// Array Used For The Keyboard Routine
bool	active=TRUE;		// Window Active Flag Set To TRUE By Default
bool	fullscreen=FALSE;	// Fullscreen Flag Set To Fullscreen Mode By Default

LRESULT	CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);	// Declaration For WndProc

// Module header, includes version, description, and address of the module retriever function
winampVisHeader hdr = { VIS_HDRVER, "Nullsoft Test Visualization Library v1.0", getModule };

// first module (oscilliscope)
winampVisModule mod1 =
{
	"Joe's VisLib",
		NULL,	// hwndParent
		NULL,	// hDllInstance
		0,		// sRate
		0,		// nCh
		10,		// latencyMS
		13,		// delayMS
		2,		// spectrumNch
		2,		// waveformNch
	{ 0, },	// spectrumData
	{ 0, },	// waveformData
	config,
	init,
	render1,
	quit
};

// second module (spectrum analyser)
/*winampVisModule mod2 =
{
"Spectrum Analyser",
NULL,	// hwndParent
NULL,	// hDllInstance
0,		// sRate
0,		// nCh
25,		// latencyMS
25,		// delayMS
2,		// spectrumNch
0,		// waveformNch
{ 0, },	// spectrumData
{ 0, },	// waveformData
config,
init,
render2, 
quit
};

  // third module (VU meter)
  winampVisModule mod3 =
  {
  "VU Meter",
  NULL,	// hwndParent
  NULL,	// hDllInstance
  0,		// sRate
  0,		// nCh
  25,		// latencyMS
  25,		// delayMS
  0,		// spectrumNch
  2,		// waveformNch
  { 0, },	// spectrumData
  { 0, },	// waveformData
  config,
  init,
  render3, 
  quit
};*/


// this is the only exported symbol. returns our main header.
// if you are compiling C++, the extern "C" { is necessary, so we just #ifdef it
#ifdef __cplusplus
extern "C" {
#endif
	__declspec( dllexport ) winampVisHeader *winampVisGetHeader()
	{
		return &hdr;
	}
#ifdef __cplusplus
}
#endif

GLuint	base;
GLYPHMETRICSFLOAT gmf[256];	// Storage For Information About Our Outline Font Characters

GLvoid BuildFont(GLvoid)								// Build Our Bitmap Font
{
	HFONT	font;										// Windows Font ID
	
	base = glGenLists(256);								// Storage For 256 Characters
	
	font = CreateFont(	-6,							// Height Of Font
		0,								// Width Of Font
		0,								// Angle Of Escapement
		0,								// Orientation Angle
		FW_NORMAL,						// Font Weight
		FALSE,							// Italic
		FALSE,							// Underline
		FALSE,							// Strikeout
		ANSI_CHARSET,					// Character Set Identifier
		OUT_TT_PRECIS,					// Output Precision
		CLIP_DEFAULT_PRECIS,			// Clipping Precision
		ANTIALIASED_QUALITY,			// Output Quality
		FF_DONTCARE|DEFAULT_PITCH,		// Family And Pitch
		//"Comic Sans MS");				// Font Name
		//"Times New Roman");				// Font Name
		"Tahoma");				// Font Name

	SelectObject(hDC, font);							// Selects The Font We Created
	
	wglUseFontOutlines(	hDC,							// Select The Current DC
		0,								// Starting Character
		255,							// Number Of Display Lists To Build
		base,							// Starting Display Lists
		10,		// Deviation From The True Outlines
		.1,			// Font Thickness In The Z Direction
		WGL_FONT_POLYGONS,				// Use Polygons, Not Lines
		gmf);							// Address Of Buffer To Recieve Data
}

GLvoid KillFont(GLvoid)									// Delete The Font
{
	glDeleteLists(base, 256);								// Delete All 256 Characters
}

GLvoid glPrint(bool bMid,const char *fmt, ...)					// Custom GL "Print" Routine
{
	glPushMatrix();
	float		length=0;								// Used To Find The Length Of The Text
	char		text[256];								// Holds Our String
	va_list		ap;										// Pointer To List Of Arguments
	
	if (fmt == NULL)									// If There's No Text
		return;											// Do Nothing
	
	va_start(ap, fmt);									// Parses The String For Variables
	vsprintf(text, fmt, ap);						// And Converts Symbols To Actual Numbers
	va_end(ap);											// Results Are Stored In Text
	
	for (unsigned int loop=0;loop<(strlen(text));loop++)	// Loop To Find Text Length
	{
		length+=gmf[text[loop]].gmfCellIncX;			// Increase Length By Each Characters Width
	}
	
	if(bMid)glTranslatef(-length/2,0.0f,0.0f);					// Center Our Text On The Screen
	
	glPushAttrib(GL_LIST_BIT);							// Pushes The Display List Bits
	glListBase(base);									// Sets The Base Character to 0
	glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);	// Draws The Display List Text
	glPopAttrib();										// Pops The Display List Bits
	glPopMatrix();
}


GLvoid ReSizeGLScene(GLsizei width, GLsizei height)		// Resize And Initialize The GL Window
{
	if (height==0)										// Prevent A Divide By Zero By
	{
		height=1;										// Making Height Equal One
	}
	
	glViewport(0,0,width,height);						// Reset The Current Viewport
	
	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glLoadIdentity();									// Reset The Projection Matrix
	
	// Calculate The Aspect Ratio Of The Window
	gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,0.1f,100.0f);
	
	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

int InitGL(GLvoid)										// All Setup For OpenGL Goes Here
{
	glShadeModel(GL_SMOOTH);							// Enable Smooth Shading
	glClearColor(0.0f, 0.0f, 0.0f, .0f);				// Black Background
	glClearAccum(0.0f, 0.0f, 0.0f, .0f);
	glClearDepth(1.0f);									// Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);							// Enables Depth Testing

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_DST_ALPHA);

	glDepthFunc(GL_LESS);								// The Type Of Depth Testing To Do
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);	// Really Nice Perspective Calculations

	dplists = glGenLists(1);
	dplists1 = glGenLists(1);

	BuildFont();

	CFTimer.Start(60);

	return TRUE;										// Initialization Went OK
}
#define _S 50.f
#define _H 2.f
#define _DIFFDPL .25f
#define _SLOWDOWN .9f
#define _SPEED .3f
#define _LTLOWER1 .92
#define _LTLOWER2 .995
#define _LTLOWER3 .999
#define _HTM 8
#define _HTM1 5

int DrawGLScene(struct winampVisModule *this_mod)									// Here's Where We Do All The Drawing
{
	int x,y,iglDrawType;
	static float fry=0,frx=0,frz=0;
	static float total = 0;
	static float frvy=0,frvx=0,frvz=0;
	static float ddx = 0;
	static float dddx = 0;
	static float xgoff,ygoff;
	static long lVertS = 0;
	float fTLOWER1,fTLOWER2,fTLOWER3;
	float fFrameCmp = 20.f / CFTimer.GetFramesPerSec();

	lFrame ++;

	iglDrawType = GL_LINE_STRIP;

	//glClear(GL_ACCUM_BUFFER_BIT);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer

	if(bDisplayFPS){
		static float ffps = CFTimer.GetFramesPerSec(),frx=0,fry=0,fs=0,r,g,b;
		static float fVertices,fCPS;
		static long ltime = time(NULL);
		if(ltime - time(NULL) < 0){
			ltime = time(NULL);
			ffps = CFTimer.GetFramesPerSec();
			fVertices = (((_X-1)*(_Y-1)*4*(2*g_iFInsts+1)*(2*g_iFInsts+1))+lVertS*(2*g_iFInsts+1)*(2*g_iFInsts+1))*ffps/1000.0;
			fCPS = _X*_Y*_DIM*_PATTERNPERINPUT*ffps/1000000.f;
		}
		glLoadIdentity();									// Reset The View
	
		glTranslatef(-.0f,3.5f,-10.f);
		
		r = ffps < 75.0 ? 1-ffps/75.0:0;
		g = ffps > 0.0 ? (ffps)/75.0f:0;
		g>1?g=1:g=g;
		b =1-r-g;
		b<0?b=0:g=g;
		glColor4f(r,g,b,1);
		
		float f[3] = {r,g,b};
		float fa[3] = {r/2.f,g/2.f,b/2.f};
		
		glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,f);
		glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,fa);
		
		glScalef(.3,.3,.3);
		
		glPrint(false,"%.1f MCPS %.1f kVerts/s - %.2f fps %2.2f",fCPS,fVertices,ffps,total);
	}

	lVertS = 0;

	total = 0;
	for (y=0; y < this_mod->nCh; y++)  // channels
	{
		for (x = 0; x < 576; x ++)
			total += abs((this_mod->waveformData[y][x]^128) - 128);
	}
	total /= 576.0;
	total /= this_mod->nCh;

	glLoadIdentity();									// Reset The Current Modelview Matrix

	//glTranslatef(-3,-3,0);

	//glScalef((total+50)/100,1,1);

	// draw minwinnerdistances
	list<nVec_var>::iterator iter = g_LWinnerDist.begin();
	glTranslatef(-5,2.5,-10);
	glColor3f(0,0,1);
	glBegin(GL_LINE_STRIP);
	glVertex3f(_WINNERDISTDS,1,0);
	glVertex3f(0,1,0);
	glEnd();
	glColor3f(1,1,1);
	glBegin(GL_LINE_STRIP);
	glVertex3f(_WINNERDISTDS,0,0);
	glVertex3f(0,0,0);
	
	float f = 0.f;
	while(iter != g_LWinnerDist.end()){
		glVertex3f(f,*iter/4,0);
		f += _WINNERDISTDS / _MAX_WINNERDIST;
		iter ++;
	};
	glEnd();
	// end - draw minwinnerdistances

	glLoadIdentity();									// Reset The Current Modelview Matrix

	glRotatef(frx,1,0,0);
	glRotatef(fry,0,1,0);
	glRotatef(frz,0,0,1);

	glTranslatef(0,0,g_fzOffset);

	if(!g_bStatic){
		frvx -= frx/5.f;
		frvy -= fry/10.f;
		frvz -= frz/500.f;
		frvx += ((this_mod->spectrumData[0][2])) / 100.f;
		frvx -= ((this_mod->spectrumData[0][10])) / 100.f;
		frvy += ((this_mod->spectrumData[0][15])) / 100.f;
		frvy -= ((this_mod->spectrumData[0][150])) / 100.f;
		frvz += ((this_mod->spectrumData[0][20])) / 40.f;
		frvz -= ((this_mod->spectrumData[0][10])) / 40.f;

		frx += frvx * _SPEED * fFrameCmp;
		fry += frvy * _SPEED * fFrameCmp;
		frz += frvz * _SPEED * fFrameCmp;

		frvx *= _SLOWDOWN * fFrameCmp;
		frvy *= _SLOWDOWN * fFrameCmp;
		frvz *= _SLOWDOWN * fFrameCmp;
	}
	else{
		frx=fry=frz = 0;
	}
	/*
	for (y = 0; y < this_mod->nCh; y ++)
	{
		glBegin(iglDrawType);
		glColor3f(0,0,255);
		for (x = 0; x < _DIM; x ++)
		{
			glVertex3f(x/_S,((y*256 + this_mod->waveformData[y][x]^128))/_S,-9);
		}
		glEnd();
	}*/

	//glTranslatef(0,0,sin(ddx) * 30 - 30);
	//ddx-=.02;
	//glTranslatef(0,0,-ddx/2);
	ddx += (total - ddx)*.2;
	dddx += (ddx - dddx)*.1;
	
	if(dddx > 10){
		fTLOWER1 = _LTLOWER1;
		fTLOWER2 = _LTLOWER2;
		fTLOWER3 = _LTLOWER3;
	}
	else{
		fTLOWER1 = _LTLOWER1 * (.8f+dddx/50.f);
		fTLOWER2 = _LTLOWER2 * (.8f+dddx/50.f);
		fTLOWER3 = _LTLOWER3 * (.8f+dddx/50.f);
	}

	iglDrawType = GL_LINE_STRIP;
	//iglDrawType = GL_TRIANGLE_STRIP;
	//iglDrawType = GL_QUAD_STRIP;

	//glTranslatef(0,0,-_MAX_LISTS * _DIFFDPL + 6);

	y=0;
	for (x = 0; x < _DIM; x ++)		// copy for som
	{
		dpInput[x] = (this_mod->spectrumData[y][x])>>(this_mod->nCh-1);
	}
	nVec tN(_DIM);
	tN.SetMemVec(dpInput);
	tN.Normalize();
	memcpy(&(pfOldData[_DIM * (lFrame % _PATTERNPERINPUT) ]),dpInput,sizeof(nVec_var) *  _DIM);
	// create "real" input pattern, i.e. the window with the _PATTERNPERINPUT size

	long l;
	for(l=0; l < _PATTERNPERINPUT; l++){
		memcpy(&(dpInput[l * _DIM]),
			&(pfOldData[_DIM*((((lFrame % _PATTERNPERINPUT)+l+1)%_PATTERNPERINPUT))]),
			sizeof(nVec_var) * _DIM);
	}

	/*FILE *fhd = fopen("c:\\testqwert.txt","w");
	for(l=0; l < _DIM*_PATTERNPERINPUT; l++){
		fprintf(fhd,"%.0f\t",dpInput[l]);
		if(!(l%_DIM))
			fprintf(fhd,"\n");
	}
	fprintf(fhd,"\n");
	fclose(fhd);*/
	
	try{
	long lx,ly,ox=0,oy=0;
	float fs = 10.f*2/_X/3,fzOffset;
	xgoff = fs * _X / 2;
	ygoff = fs * _Y / 2;

	/*nVec tN(_DIM * _PATTERNPERINPUT);
	tN.SetMemVec(dpInput);
	tN.Normalize();*/

	/*FILE *fhd = fopen("c:\\testqwert.txt","a");
	for(lx=0;lx<_X;lx++)for(ly=0;ly<_Y;ly++)
	{
		for(l=0; l < _DIM*_PATTERNPERINPUT; l++){fprintf(fhd,"%.0f\t",gSOM.GetNWeights(lx,ly)[l]);}
		fprintf(fhd,"\n");
	}
	fclose(fhd);*/

	gSOM.SetInput(dpInput);

	gSOM.GetWinner(lx,ly);

	// update the graph data for the WinnerMinDist
	g_LWinnerDist.push_back(gSOM.getWinnerMin());
	if(g_LWinnerDist.size() > _MAX_WINNERDIST){
		g_LWinnerDist.erase(g_LWinnerDist.begin());
	}

	for(x=0;x < _X;x ++){
		for(y=0;y < _Y;y ++){
			pdLT1[x][y] *= pow(fTLOWER1,fFrameCmp);
			pdLT2[x][y] *= pow(fTLOWER2,fFrameCmp);
			pdLT3[x][y] *= pow(fTLOWER3,fFrameCmp);
		}
	}
	pdLT1[lx][ly] += .8;
	if(pdLT1[lx][ly]>1){
		pdLT1[lx][ly] = 1;
	}
	pdLT2[lx][ly] += .3;
	if(pdLT2[lx][ly]>1){
		pdLT2[lx][ly] = 1;
	}
	pdLT3[lx][ly] += .3;
	if(pdLT3[lx][ly]>1){
		pdLT3[lx][ly] = 1;
	}

	for(x=0;x < _X;x ++){				// calc colors
		for(y=0;y < _Y;y ++){
			fzOffset = pdLT3[x][y] * _HTM1 + pdLT1[x][y] * 3;
			
			pppfVertex2[x][y][0] = (x)*fs+.5*fs - xgoff;
			pppfVertex2[x][y][1] = (y)*fs+.5*fs - ygoff;
			pppfVertex2[x][y][2] = -16 + fzOffset;
			
			if(x == lx && y == ly){
				pppfColor2[x][y][0] = 1;
				pppfColor2[x][y][1] = 1;
				pppfColor2[x][y][2] = 1;
			}
			else{
				pppfColor2[x][y][0] = pow(pdLT2[x][y],(255-this_mod->spectrumData[0][6])/255.0f)*(pdLT1[x][y]);
				pppfColor2[x][y][1] = pow(/*pdLT1[x][y]*/fzOffset,(255-this_mod->spectrumData[0][3])/255.0f)*
					(1.f-pdLT3[x][y])*pdLT2[x][y]*(1.f-pppfColor2[x][y][0])*total/25.f;
				pppfColor2[x][y][2] = (pdLT1[x][y])*(total/40.0f+ddx/40.f)*(1-pppfColor2[x][y][0]);
			}
		}
	}

	glNewList(dplists,GL_COMPILE);
		if(!bwires)
			glBegin(GL_QUADS);
		else 
			glBegin(GL_LINE_LOOP);
		for(x=0;x < _X;x ++){
			for(y=0;y < _Y;y ++){
				if(bwires){
					glEnd();
					glBegin(GL_LINE_LOOP);
				}
				
				fzOffset = pdLT1[x][y] * _HTM + pdLT2[x][y] * 3;
				
				fzOffset /= 2.f;
				
				if(fzOffset < .1)
					continue;
				
				if(x == lx && y == ly){
					glColor3f(pdLT1[x][y],pdLT2[x][y],1);
				}
				else{
					//glColor3f(pow(pdLT1[x][y],(100.f-total)/100.0f),pow(pdLT2[x][y],(100.f-total)/100.0f),total/80.0f);
					glColor3f(pow(pdLT1[x][y],(255-this_mod->spectrumData[0][2])/255.0f)*(1.f-pdLT1[x][y])*pdLT1[x][y]*4,
						pow(pdLT1[x][y],(255-this_mod->spectrumData[0][5])/255.0f)+(pdLT2[x][y])*(pdLT2[x][y]),
						fzOffset*(total/40.0f+ddx/40.f));
				}
				
				glVertex3f((x)*fs - xgoff,
					(y)*fs - ygoff,
					-15 + fzOffset);
				glVertex3f((x+1)*fs - xgoff,
					(y)*fs - ygoff,
					-15 + fzOffset);
				glVertex3f((x+1)*fs - xgoff,
					(y+1)*fs - ygoff,
					-15 + fzOffset);
				glVertex3f((x)*fs - xgoff,
					(y+1)*fs - ygoff,
					-15 + fzOffset);

				lVertS += 4;
			}
		}
	glEnd();
	glEndList();
	int yo,xo;
	glNewList(dplists1,GL_COMPILE);
		for(x=0;x < _X-1;x ++){
			if(!bwires)
				glBegin(GL_TRIANGLE_STRIP);
			else 
				glBegin(GL_LINE_STRIP);
			for(y=0;y < _Y-1;y ++){
				for(xo=0;xo<=1;xo++){
					for(yo=0;yo<=1;yo++){
						if(x == lx && y == ly){
							glColor3f(pdLT1[x+xo][y+yo],pdLT2[x+xo][y+yo],1);
						}
						else{
							//glColor3f(pow(pdLT1[x][y],(100.f-total)/100.0f),pow(pdLT2[x][y],(100.f-total)/100.0f),total/80.0f);
							glColor3f(
								pppfColor2[x+xo][y+yo][0],
								pppfColor2[x+xo][y+yo][1],
								pppfColor2[x+xo][y+yo][2]
								);
						}
						
						glVertex3f(
							pppfVertex2[x+xo][y+yo][0],
							pppfVertex2[x+xo][y+yo][1],
							pppfVertex2[x+xo][y+yo][2]
							);
					}
				}
			}
			glEnd();
		}
	glEndList();
	/*glNewList(dplists1,GL_COMPILE);
	glBegin(GL_TRIANGLE_STRIP);
	for(x=0;x < _X-1;x ++){
	glBegin(GL_TRIANGLE_STRIP);
			for(y=0;y < _Y-1;y ++){
				fzOffset = pdLT3[x][y] * _HTM1 + pdLT1[x][y] * 3;
				
				if(x == lx && y == ly){
					glColor3f(pdLT1[x][y],pdLT2[x][y],1);
				}
				else{
					//glColor3f(pow(pdLT1[x][y],(100.f-total)/100.0f),pow(pdLT2[x][y],(100.f-total)/100.0f),total/80.0f);
					glColor3f(pow(pdLT1[x][y],(255-this_mod->spectrumData[0][2])/255.0f)*(1.f-fzOffset)*fzOffset*4,
						pow(pdLT1[x][y],(255-this_mod->spectrumData[0][5])/255.0f)+(fzOffset)*(fzOffset),
						fzOffset*(total/40.0f+ddx/40.f));
				}

				glVertex3f((x)*fs,(y)*fs,-15 + fzOffset);
				glVertex3f((x+1)*fs,(y)*fs,-15 + fzOffset);
				glVertex3f((x)*fs,(y+1)*fs,-15 + fzOffset);
				glVertex3f((x+1)*fs,(y+1)*fs,-15 + fzOffset);
			}
			glEnd();
		}
		glEnd();
	glEndList();*/

	float fXxs = _X * fs;
	float fYys = _Y * fs;
	for(ox=-g_iFInsts;ox<=g_iFInsts;ox++){
		for(oy=-g_iFInsts;oy<=g_iFInsts;oy++){
			glPushMatrix();
			glTranslatef(ox*fXxs,oy*fYys,0);
			glCallList(dplists);
			glPopMatrix();
		}
	}
	for(ox=-g_iFInsts;ox<=g_iFInsts;ox++){
		for(oy=-g_iFInsts;oy<=g_iFInsts;oy++){
			glPushMatrix();
			glTranslatef(ox*fXxs,oy*fYys,0);
			if(!g_bStatic)
				glScalef(0.4f+ddx/60.f,0.4f+ddx/60.f,1.f);
			glCallList(dplists1);
			glPopMatrix();
		}
	}

	//glAccum(GL_ACCUM,1.f);

	//glAccum(GL_RETURN,1.f);
	//glFlush();

	/*static float frax=0;
	frax += (50-total) * .01;
	glTranslatef(0,total/10.f,0);

	glRotatef((total - .5)*.2,1,0,0);
	//glRotatef(frax,0,0,1);
glScalef(1,-(total+70)/100,1);
glTranslatef(0,-total/10,0);
	for(ox=-g_iFInsts;ox<=g_iFInsts;ox++){
		for(oy=-g_iFInsts;oy<=g_iFInsts;oy++){
			glPushMatrix();
			glTranslatef(ox*fXxs,oy*fYys,-5);
			glScalef(2,2,2);
			glCallList(dplists);
			glPopMatrix();
		}
	}
	glRotatef(frax,0,0,1);
	for(ox=-g_iFInsts;ox<=g_iFInsts;ox++){
		for(oy=-g_iFInsts;oy<=g_iFInsts;oy++){
			glPushMatrix();
			glTranslatef(ox*fXxs,oy*fYys,-3);
			//glScalef(.5,.5,.5);
			glCallList(dplists);
			glPopMatrix();
		}
	}
	glTranslatef(10,1,2+ddx/4.5);
	glRotatef(frax,0,0,1);
	glTranslatef(-10,-1,-2);
	glRotatef(total,0,1,0);
	for(ox=-g_iFInsts;ox<=g_iFInsts;ox++){
		for(oy=-g_iFInsts;oy<=g_iFInsts;oy++){
			glPushMatrix();
			glTranslatef(ox*fXxs,oy*fYys,-2);
			glCallList(dplists);
			glPopMatrix();
		}
	}*/

	}
	catch(...){
	}

	return TRUE;										// Everything Went OK
}

GLvoid KillGLWindow(GLvoid)								// Properly Kill The Window
{
	if (fullscreen)										// Are We In Fullscreen Mode?
	{
		ChangeDisplaySettings(NULL,0);					// If So Switch Back To The Desktop
		ShowCursor(TRUE);								// Show Mouse Pointer
	}
	
	if (hRC)											// Do We Have A Rendering Context?
	{
		if (!wglMakeCurrent(NULL,NULL))					// Are We Able To Release The DC And RC Contexts?
		{
			MessageBox(NULL,"Release Of DC And RC Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}
		
		if (!wglDeleteContext(hRC))						// Are We Able To Delete The RC?
		{
			MessageBox(NULL,"Release Rendering Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}
		hRC=NULL;										// Set RC To NULL
	}
	
	if (hDC && !ReleaseDC(hWnd,hDC))					// Are We Able To Release The DC
	{
		MessageBox(NULL,"Release Device Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hDC=NULL;										// Set DC To NULL
	}
	
	if (hWnd && !DestroyWindow(hWnd))					// Are We Able To Destroy The Window?
	{
		MessageBox(NULL,"Could Not Release hWnd.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hWnd=NULL;										// Set hWnd To NULL
	}
	
	if (!UnregisterClass("OpenGL",hInstance))			// Are We Able To Unregister Class
	{
		MessageBox(NULL,"Could Not Unregister Class.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hInstance=NULL;									// Set hInstance To NULL
	}
	KillFont();
}

LRESULT CALLBACK WndProc(	HWND	hWnd,			// Handle For This Window
						 UINT	uMsg,			// Message For This Window
						 WPARAM	wParam,			// Additional Message Information
						 LPARAM	lParam)			// Additional Message Information
{
	switch (uMsg)									// Check For Windows Messages
	{
	case WM_ACTIVATE:							// Watch For Window Activate Message
		{
			if (!HIWORD(wParam))					// Check Minimization State
			{
				active=TRUE;						// Program Is Active
			}
			else
			{
				active=FALSE;						// Program Is No Longer Active
			}
			
			return 0;								// Return To The Message Loop
		}
		
	case WM_SYSCOMMAND:							// Intercept System Commands
		{
			switch (wParam)							// Check System Calls
			{
			case SC_SCREENSAVE:					// Screensaver Trying To Start?
			case SC_MONITORPOWER:				// Monitor Trying To Enter Powersave?
				return 0;							// Prevent From Happening
			}
			break;									// Exit
		}
		
	case WM_CLOSE:								// Did We Receive A Close Message?
		{
			PostQuitMessage(0);						// Send A Quit Message
			
			return 0;								// Jump Back
		}
		
	case WM_KEYDOWN:							// Is A Key Being Held Down?
		{
			// get this_mod from our window's user data
			winampVisModule *this_mod = (winampVisModule *) GetWindowLong(hWnd,GWL_USERDATA);
			PostMessage(this_mod->hwndParent,uMsg,wParam,lParam);
			
			keys[wParam] = TRUE;					// If So, Mark It As TRUE
			if(wParam == 'W'){
				bwires =! bwires;
				if(bwires){
					glClearColor(0.5f, 0.5f, 0.5f, .0f);
				}
				else{
					glClearColor(0.0f, 0.0f, 0.0f, .0f);
				}
			}
			if(wParam == 'A'){
				g_bStatic = !g_bStatic;
			}
			if(wParam == 'S'){
				if(g_iFInsts > 0)
					g_iFInsts --;
			}
			if(keys['1']){
				g_fzOffset += .4;
			}
			if(keys['2']){
				g_fzOffset -= .4;
			}
			if(wParam == 'D'){
				g_iFInsts ++;
			}
			if(wParam == 'F'){
				bSwitchMode = true;
			}
			if(wParam == 'G'){
				bDisplayFPS =! bDisplayFPS;
			}
			if(wParam == ' '){
				int x,y;
				for(x=0;x < _X;x ++){
					for(y=0;y < _Y;y ++){
						pdLT1[x][y] = 0;
						pdLT2[x][y] = 0;
						pdLT3[x][y] = 0;
					}
				}
			}
			if(wParam == 'R'){
				gSOM.InitWeights(Init_Random);
				int x,y;
				for(x=0;x < _X;x ++){
					for(y=0;y < _Y;y ++){
						pdLT1[x][y] = 0;
						pdLT2[x][y] = 0;
						pdLT3[x][y] = 0;
					}
				}
			}
			return 0;								// Jump Back
		}
		
	case WM_KEYUP:								// Has A Key Been Released?
		{
			// get this_mod from our window's user data
			winampVisModule *this_mod = (winampVisModule *) GetWindowLong(hWnd,GWL_USERDATA);
			PostMessage(this_mod->hwndParent,uMsg,wParam,lParam);
			
			keys[wParam] = FALSE;					// If So, Mark It As FALSE
			return 0;								// Jump Back
		}
		
	case WM_SIZE:								// Resize The OpenGL Window
		{
			ReSizeGLScene(LOWORD(lParam),HIWORD(lParam));  // LoWord=Width, HiWord=Height
			return 0;								// Jump Back
		}
	}
	
	// Pass All Unhandled Messages To DefWindowProc
	return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

/*int WINAPI WinMain(	HINSTANCE	hInstance,			// Instance
				   HINSTANCE	hPrevInstance,		// Previous Instance
				   LPSTR		lpCmdLine,			// Command Line Parameters
				   int			nCmdShow)			// Window Show State
{
	MSG		msg;									// Windows Message Structure
	BOOL	done=FALSE;								// Bool Variable To Exit Loop
	
	// Ask The User Which Screen Mode They Prefer
	if (MessageBox(NULL,"Would You Like To Run In Fullscreen Mode?", "Start FullScreen?",MB_YESNO|MB_ICONQUESTION)==IDNO)
	{
		fullscreen=FALSE;							// Windowed Mode
	}
	
	// Create Our OpenGL Window
	if (!CreateGLWindow("NeHe's OpenGL Framework",640,480,16,fullscreen))
	{
		return 0;									// Quit If Window Was Not Created
	}
	
	while(!done)									// Loop That Runs While done=FALSE
	{
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))	// Is There A Message Waiting?
		{
			if (msg.message==WM_QUIT)				// Have We Received A Quit Message?
			{
				done=TRUE;							// If So done=TRUE
			}
			else									// If Not, Deal With Window Messages
			{
				TranslateMessage(&msg);				// Translate The Message
				DispatchMessage(&msg);				// Dispatch The Message
			}
		}
		else										// If There Are No Messages
		{
			// Draw The Scene.  Watch For ESC Key And Quit Messages From DrawGLScene()
			if (active)								// Program Active?
			{
				if (keys[VK_ESCAPE])				// Was ESC Pressed?
				{
					done=TRUE;						// ESC Signalled A Quit
				}
				else								// Not Time To Quit, Update Screen
				{
					DrawGLScene();					// Draw The Scene
					SwapBuffers(hDC);				// Swap Buffers (Double Buffering)
				}
			}
			
			if (keys[VK_F1])						// Is F1 Being Pressed?
			{
				keys[VK_F1]=FALSE;					// If So Make Key FALSE
				KillGLWindow();						// Kill Our Current Window
				fullscreen=!fullscreen;				// Toggle Fullscreen / Windowed Mode
				// Recreate Our OpenGL Window
				if (!CreateGLWindow("NeHe's OpenGL Framework",640,480,16,fullscreen))
				{
					return 0;						// Quit If Window Was Not Created
				}
			}
		}
	}
	
	// Shutdown
	KillGLWindow();									// Kill The Window
	return (msg.wParam);							// Exit The Program
}*/


// getmodule routine from the main header. Returns NULL if an invalid module was requested,
// otherwise returns either mod1, mod2 or mod3 depending on 'which'.
winampVisModule *getModule(int which)
{
	switch (which)
	{
	case 0: return &mod1;
	/*case 1: return &mod2;
	case 2: return &mod3;*/
	default:return NULL;
	}
}

// configuration. Passed this_mod, as a "this" parameter. Allows you to make one configuration
// function that shares code for all your modules (you don't HAVE to use it though, you can make
// config1(), config2(), etc...)
void config(struct winampVisModule *this_mod)
{
	MessageBox(this_mod->hwndParent,"blabla","blabla",MB_OK);
}

// initialization. Registers our window class, creates our window, etc. Again, this one works for
// both modules, but you could make init1() and init2()...
// returns 0 on success, 1 on failure.
int init(struct winampVisModule *this_mod)
{
	int width = 800;
	int height = 600;
	int bits = 32;

	int fs_width = 1280;
	int fs_height = 1024;
	int fs_bits = 32;

	char title[]="blabla";

	config_read(this_mod);

	GLuint		PixelFormat;			// Holds The Results After Searching For A Match
	WNDCLASS	wc;						// Windows Class Structure
	DWORD		dwExStyle;				// Window Extended Style
	DWORD		dwStyle;				// Window Style
	RECT		WindowRect;				// Grabs Rectangle Upper Left / Lower Right Values
	if(fullscreen){
		WindowRect.left=(long)0;			// Set Left Value To 0
		WindowRect.right=(long)fs_width;		// Set Right Value To Requested Width
		WindowRect.top=(long)0;				// Set Top Value To 0
		WindowRect.bottom=(long)fs_height;		// Set Bottom Value To Requested Height
	}
	else{
		WindowRect.left=(long)0;			// Set Left Value To 0
		WindowRect.right=(long)width;		// Set Right Value To Requested Width
		WindowRect.top=(long)0;				// Set Top Value To 0
		WindowRect.bottom=(long)height;		// Set Bottom Value To Requested Height
	}
	
	//fullscreen=fullscreenflag;			// Set The Global Fullscreen Flag
	
	hInstance			= this_mod->hDllInstance;				// Grab An Instance For Our Window
	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;	// Redraw On Size, And Own DC For Window.
	wc.lpfnWndProc		= (WNDPROC) WndProc;					// WndProc Handles Messages
	wc.cbClsExtra		= 0;									// No Extra Window Data
	wc.cbWndExtra		= 0;									// No Extra Window Data
	wc.hInstance		= this_mod->hDllInstance;							// Set The Instance
	wc.hIcon			= LoadIcon(NULL, IDI_WINLOGO);			// Load The Default Icon
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);			// Load The Arrow Pointer
	wc.hbrBackground	= NULL;									// No Background Required For GL
	wc.lpszMenuName		= NULL;									// We Don't Want A Menu
	wc.lpszClassName	= "OpenGL";								// Set The Class Name
	
	if (!RegisterClass(&wc))									// Attempt To Register The Window Class
	{
		MessageBox(NULL,"Failed To Register The Window Class.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;											// Return FALSE
	}
	
	if (fullscreen)												// Attempt Fullscreen Mode?
	{
		DEVMODE dmScreenSettings;								// Device Mode
		memset(&dmScreenSettings,0,sizeof(dmScreenSettings));	// Makes Sure Memory's Cleared
		dmScreenSettings.dmSize=sizeof(dmScreenSettings);		// Size Of The Devmode Structure
		dmScreenSettings.dmPelsWidth	= fs_width;				// Selected Screen Width
		dmScreenSettings.dmPelsHeight	= fs_height;				// Selected Screen Height
		dmScreenSettings.dmBitsPerPel	= fs_bits;					// Selected Bits Per Pixel
		dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;
		
		// Try To Set Selected Mode And Get Results.  NOTE: CDS_FULLSCREEN Gets Rid Of Start Bar.
		if (ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL)
		{
			// If The Mode Fails, Offer Two Options.  Quit Or Use Windowed Mode.
			if (MessageBox(NULL,"The Requested Fullscreen Mode Is Not Supported By\nYour Video Card. Use Windowed Mode Instead?","NeHe GL",MB_YESNO|MB_ICONEXCLAMATION)==IDYES)
			{
				fullscreen=FALSE;		// Windowed Mode Selected.  Fullscreen = FALSE
			}
			else
			{
				// Pop Up A Message Box Letting User Know The Program Is Closing.
				MessageBox(NULL,"Program Will Now Close.","ERROR",MB_OK|MB_ICONSTOP);
				return FALSE;									// Return FALSE
			}
		}
	}
	
	if (fullscreen)												// Are We Still In Fullscreen Mode?
	{
		dwExStyle=WS_EX_APPWINDOW;								// Window Extended Style
		dwStyle=WS_POPUP;										// Windows Style
		ShowCursor(FALSE);										// Hide Mouse Pointer
	}
	else
	{
		dwExStyle=WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;			// Window Extended Style
		dwStyle=WS_OVERLAPPEDWINDOW;							// Windows Style
	}
	
	AdjustWindowRectEx(&WindowRect, dwStyle, FALSE, dwExStyle);		// Adjust Window To True Requested Size
	
	// Create The Window
	if (!(hWnd=CreateWindowEx(	dwExStyle,							// Extended Style For The Window
								"OpenGL",							// Class Name
								title,								// Window Title
								dwStyle |							// Defined Window Style
								WS_VISIBLE|WS_SYSMENU,					// Required Window Style
								0, 0,								// Window Position
								WindowRect.right-WindowRect.left,	// Calculate Window Width
								WindowRect.bottom-WindowRect.top,	// Calculate Window Height
								this_mod->hwndParent,								// No Parent Window
								NULL,								// No Menu
								this_mod->hDllInstance,							// Instance
								NULL)))								// Dont Pass Anything To WM_CREATE
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Window Creation Error.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	SetWindowLong(hWnd,GWL_USERDATA,(LONG)this_mod); // set our user data to a "this" pointer
	
	static	PIXELFORMATDESCRIPTOR pfd=				// pfd Tells Windows How We Want Things To Be
	{
		sizeof(PIXELFORMATDESCRIPTOR),				// Size Of This Pixel Format Descriptor
			1,											// Version Number
			PFD_DRAW_TO_WINDOW |						// Format Must Support Window
			PFD_SUPPORT_OPENGL |						// Format Must Support OpenGL
			PFD_DOUBLEBUFFER,							// Must Support Double Buffering
			PFD_TYPE_RGBA,								// Request An RGBA Format
			bits,										// Select Our Color Depth
			0, 0, 0, 0, 0, 0,							// Color Bits Ignored
			0,											// No Alpha Buffer
			0,											// Shift Bit Ignored
			0,											// No Accumulation Buffer
			0, 0, 0, 0,									// Accumulation Bits Ignored
			16,											// 16Bit Z-Buffer (Depth Buffer)  
			0,											// No Stencil Buffer
			0,											// No Auxiliary Buffer
			PFD_MAIN_PLANE,								// Main Drawing Layer
			0,											// Reserved
			0, 0, 0										// Layer Masks Ignored
	};
	
	if (!(hDC=GetDC(hWnd)))							// Did We Get A Device Context?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Create A GL Device Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}
	
	if (!(PixelFormat=ChoosePixelFormat(hDC,&pfd)))	// Did Windows Find A Matching Pixel Format?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Find A Suitable PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}
	
	if(!SetPixelFormat(hDC,PixelFormat,&pfd))		// Are We Able To Set The Pixel Format?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Set The PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}
	
	if (!(hRC=wglCreateContext(hDC)))				// Are We Able To Get A Rendering Context?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Create A GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}
	
	if(!wglMakeCurrent(hDC,hRC))					// Try To Activate The Rendering Context
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Activate The GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}
	
	ShowWindow(hWnd,SW_SHOW);						// Show The Window
	SetForegroundWindow(hWnd);						// Slightly Higher Priority
	SetFocus(hWnd);									// Sets Keyboard Focus To The Window
	if(fullscreen){
		ReSizeGLScene(fs_width, fs_height);					// Set Up Our Perspective GL Screen
	}
	else{
		ReSizeGLScene(width, height);					// Set Up Our Perspective GL Screen
	}
	
	if (!InitGL())									// Initialize Our Newly Created GL Window
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Initialization Failed.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	DrawGLScene(this_mod);
	SwapBuffers(hDC);
	// init som shit
	int x,y;
	for(x=0;x < _X;x ++){
		for(y=0;y < _Y;y ++){
			pdLT1[x][y] = 0;
			pdLT2[x][y] = 0;
		}
	}

//	SP.SetMaxPatternNum(10000);
	gSOM.SetSize(_DIM*_PATTERNPERINPUT,_X,_Y);
	gSOM.InitWeights(Init_Random);
	gSOM.Load("c:\\sound.som");
	//gSOM.Save("c:\\sound.som");
	//gSOM.InitWeights(Init_Random);
	return 0;
}

// render function for oscilliscope. Returns 0 if successful, 1 if visualization should end.
int render1(struct winampVisModule *this_mod)
{
	if(bSwitchMode){
		bSwitchMode = false;
		
		if (fullscreen)										// Are We In Fullscreen Mode?
		{
			ChangeDisplaySettings(NULL,0);					// If So Switch Back To The Desktop
			ShowCursor(TRUE);								// Show Mouse Pointer
		}

		fullscreen = !fullscreen;

		quit(this_mod);
		init(this_mod);
	}
	CFTimer.Frame();
	/*int x, y;
	// clear background
	Rectangle(memDC,0,0,_DIM,256);
	// draw oscilliscope
	for (y = 0; y < this_mod->nCh; y ++)
	{
		MoveToEx(memDC,0,(y*256)>>(this_mod->nCh-1),NULL);
		for (x = 0; x < _DIM; x ++)
		{
			LineTo(memDC,x,(y*256 + this_mod->waveformData[y][x]^128)>>(this_mod->nCh-1));
		}
	}
	{ // copy doublebuffer to window
		HDC hdc = GetDC(hMainWnd);
		BitBlt(hdc,0,0,_DIM,256,memDC,0,0,SRCCOPY);
		ReleaseDC(hMainWnd,hdc);
	}*/
	HDC hdc = GetDC(hWnd);
	DrawGLScene(this_mod);
	SwapBuffers(hdc);
	return 0;
}
/*
// render function for analyser. Returns 0 if successful, 1 if visualization should end.
int render2(struct winampVisModule *this_mod)
{
int x, y;
// clear background
Rectangle(memDC,0,0,_DIM,256);
// draw analyser
for (y = 0; y < this_mod->nCh; y ++)
{
for (x = 0; x < _DIM; x ++)
{
MoveToEx(memDC,x,(y*256+256)>>(this_mod->nCh-1),NULL);
LineTo(memDC,x,(y*256 + 256 - this_mod->spectrumData[y][x])>>(this_mod->nCh-1));
}
}
{ // copy doublebuffer to window
HDC hdc = GetDC(hMainWnd);
BitBlt(hdc,0,0,_DIM,256,memDC,0,0,SRCCOPY);
ReleaseDC(hMainWnd,hdc);
}
return 0;
}

  // render function for VU meter. Returns 0 if successful, 1 if visualization should end.
  int render3(struct winampVisModule *this_mod)
  {
  int x, y;
  // clear background
  Rectangle(memDC,0,0,256,32);
  // draw VU meter
  for (y = 0; y < 2; y ++)
  {
		int last=this_mod->waveformData[y][0];
		int total=0;
		for (x = 1; x < 576; x ++)
		{
		total += abs(last - this_mod->waveformData[y][x]);
		last = this_mod->waveformData[y][x];
		}
		total /= _DIM;
		if (total > 127) total = 127;
		if (y) Rectangle(memDC,128,0,128+total,32);
		else Rectangle(memDC,128-total,0,128,32);
		}
		{ // copy doublebuffer to window
		HDC hdc = GetDC(hMainWnd);
		BitBlt(hdc,0,0,256,32,memDC,0,0,SRCCOPY);
		ReleaseDC(hMainWnd,hdc);
		}
		return 0;
		}
*/
// cleanup (opposite of init()). Destroys the window, unregisters the window class
void quit(struct winampVisModule *this_mod)
{
	//SP.Save("c:\\sound.spt");

	config_write(this_mod);		// write configuration
/*	SelectObject(memDC,oldBM);	// delete our doublebuffer
	DeleteObject(memDC);
	DeleteObject(memBM);	*/
	KillGLWindow();
	DestroyWindow(hMainWnd); // delete our window
	UnregisterClass(szAppName,this_mod->hDllInstance); // unregister window class
}


// window procedure for our window
/*LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
switch (message)
{
case WM_CREATE:		return 0;
case WM_ERASEBKGND: return 0;
case WM_PAINT:
{ // update from doublebuffer
PAINTSTRUCT ps;
RECT r;
HDC hdc = BeginPaint(hwnd,&ps);
GetClientRect(hwnd,&r);
BitBlt(hdc,0,0,r.right,r.bottom,memDC,0,0,SRCCOPY);
EndPaint(hwnd,&ps);
}
return 0;
case WM_DESTROY: PostQuitMessage(0); return 0;
case WMg_iFInstsEYDOWN: // pass keyboard messages to main winamp window (for processing)
case WMg_iFInstsEYUP:
{	// get this_mod from our window's user data
winampVisModule *this_mod = (winampVisModule *) GetWindowLong(hwnd,GWL_USERDATA);
PostMessage(this_mod->hwndParent,message,wParam,lParam);
}
return 0;
case WM_MOVE:
{	// get config_x and config_y for configuration
RECT r;
GetWindowRect(hMainWnd,&r);
config_x = r.left;
config_y = r.top;
}
return 0;
}
return DefWindowProc(hwnd,message,wParam,lParam);
}*/


void config_getinifn(struct winampVisModule *this_mod, char *ini_file)
{	// makes a .ini file in the winamp directory named "plugin.ini"
	char *p;
	GetModuleFileName(this_mod->hDllInstance,ini_file,MAX_PATH);
	p=ini_file+strlen(ini_file);
	while (p >= ini_file && *p != '\\') p--;
	if (++p >= ini_file) *p = 0;
	strcat(ini_file,"plugin.ini");
}


void config_read(struct winampVisModule *this_mod)
{
	char ini_file[MAX_PATH];
	config_getinifn(this_mod,ini_file);
	config_x = GetPrivateProfileInt(this_mod->description,"Screen_x",config_x,ini_file);
	config_y = GetPrivateProfileInt(this_mod->description,"Screen_y",config_y,ini_file);
}

void config_write(struct winampVisModule *this_mod)
{
	char string[32];
	char ini_file[MAX_PATH];
	
	config_getinifn(this_mod,ini_file);
	
	wsprintf(string,"%d",config_x);
	WritePrivateProfileString(this_mod->description,"Screen_x",string,ini_file);
	wsprintf(string,"%d",config_y);
	WritePrivateProfileString(this_mod->description,"Screen_y",string,ini_file);
}
